#!/usr/bin/env python
# coding: utf-8


from .CRF    import CRF
from .NER    import NER
from .Tagger import Tagger
from .Multi_Head_Attention import Multi_Head_Attention
from .Transformer_Encoder import Transformer_Encoder
from .utils  import save, load
from .TextMatcher import TextMatcher
from .utils import golden_section
from .utils import positional_embeddings






import pickle
from .NER import NER
import math
def save(path, model):
    """
    Save the model parameters to a file
    inputs:
        path (str)-----> the path to save the model must be the full path + name of file without any extension
        model (NER)----> the model to save
    """
    container = {"k":model.k, "units":model.units,\
                 "m":model.m,"n":model.n,\
                 "model_weights":model.get_weights()}
    with open(str(path)+".pkl",'wb') as file:
        pickle.dump(container, file)
        print("model parameters has been saved !")


def load(path):
    """
    load the model parameters from a file
    input:
        path (str)-----> the path to the model must be the full path + name of file without any extension
    output:
        model (NER)----> the model with the loaded parameters
    """

    with open(str(path)+".pkl", 'rb') as file:
        container = pickle.load(file)
    model = NER(container["units"], container["k"], container["n"], container["m"])
    model.lstm.build((None, container["n"], container["m"]))
    model.crf.build((None, container["n"], container["units"]*2))
    model.set_weights(container["model_weights"])
    print("model weights has been loaded!")
    return model

def golden_section(fun, inte, ite = 10):
    """
    this function is used to find the minimum of a function in a given interval
    using the golden section search algorithm.
    inputs:
        fun (fun): the function to minimize
        inte (tuple): the interval to search in
        ite (int): the number of iterations
    output:
        res (float): the minimum of the function
    """
    lower,upper = inte
    r   = (math.sqrt(5) - 1)/2
    d   = r * (upper - lower)
    x1  = lower + d
    x2  = upper - d 
    counter = 1
    while(counter != ite):
        if fun(x1) > fun(x2):
            upper = x1
            x1    = x2
            d     = r * (upper - lower)
            x2    = upper - d
        else :
            lower = x2
            x2    = x1
            d     = r * (upper - lower)
            x1    = lower + d
        counter += 1
    res = (x1+x2)/2
    return res

def positional_embeddings(d, n, c = 10000):
    """
    get the positional embeddings of embeddings of size [d,n]
    inputs :
        d (int) -------------------> is the diemnsion of the embeddings
        n (int) -------------------> is the number of tokens in the sentence
        c (int) -------------------> hyperparameter by default 10000
    output :
        tokens (List[list[]])------> positional embeddings for the tokens as list of lists
    """
    tokens = []
    for position in range(n):
        tokens.append([])
        for i in range(d):
            # if i is even calculate even_pe
            if i%2 == 0:
                even_d    = math.pow(c, i/d)
                even_pe = math.sin(position/ even_d)
                tokens[position].append(even_pe)
            # if i is odd calculate odd_pe    
            else:
                odd_d    = math.pow(c, i/d)
                odd_pe = math.cos(position/ odd_d)
                tokens[position].append(odd_pe)            
    return tokens


import tensorflow as tf
from .Multi_Head_Attention import Multi_Head_Attention

class Transformer_Encoder(tf.keras.layers.Layer):
    """
    Example
    --------
        # assume we a dataset of sentence each token either a device name or undefined token
        # assume that the dataset has 10,000 setences and the maximum length of the sentence is 10
        # our goal is to build a Encoder transformer model to tag the device names in the sentence
        # first use random inputs with embedding size of 3000
        inputs = tf.random.uniform((10000, 300, 10), -1, 1)
        tags   = tf.random.uniform((10000, 10), 0, 1, dtype = tf.int32)
        # build the model
        d = 300 
        n = 10
        h = 2
        m = 1
        input_  = Input((d,n))
        encoder = Transformer_Encoder(d, n, h, m)(input_)
        pooling = tf.keras.layers.GlobalAveragePooling1D()(encoder)
        output  = Dense(n, activation = "sigmoid")(pooling)
        model   = Model(input_, output)
        model.compile(optimizer="adam",
                      loss = "binary_crossentropy",
                      metrics=["binary_accuracy"])
        # now you can train the model
        #model.fit(X, Y, epochs=30, batch_size = 2**10, validation_split=0.2)
        
        # don't use this approach unless you have a large dataset
    """
    def __init__(self, d, n, h, m, **kwargs):
        """
        Transformer-Encoder Layer
        inputs:
            d (int)----> number of features in the word embedding
            n (int)----> number of words in the sentence (maximum length)
            h (int)----> number of heads
            m (int)----> number of recursive calls to the layer (Recursive Transformer)
        """
        self.d = d
        self.n = n
        self.h = h
        self.m = m
        # transformer encoder layers
        # in future updates the user will be able to customize these layers
        self.mhsa          = Multi_Head_Attention(self.d, self.n, self.h)
        self.add_normalize = tf.keras.layers.LayerNormalization(axis = -1, epsilon=0.01)
        self.NN_layer1     = tf.keras.layers.Dense(4*self.d, activation = "relu")
        self.NN_layer2     = tf.keras.layers.Dense(self.n, activation = "relu")
        self.dropout       = tf.keras.layers.Dropout(0.5)
        super(Transformer_Encoder, self).__init__(**kwargs)
    def build(self, input_shape):
        # intialize the parameters of each layer
        self.mhsa.build(input_shape)
        self.NN_layer1.build(input_shape)
        self.NN_layer2.build(self.NN_layer1.compute_output_shape(input_shape))
        super(Transformer_Encoder, self).build(input_shape)
    def call(self, X):
        for _ in range(self.m):
            mhsa          = self.mhsa(X)
            add_normalize = self.add_normalize(X + mhsa)
            NN_layer1     = self.NN_layer1(add_normalize)
            NN_layer2     = self.NN_layer2(NN_layer1) 
            output        = self.dropout(NN_layer2 + add_normalize)
            X             = output
        return output   


import tensorflow as tf
class Multi_Head_Attention(tf.keras.layers.Layer):
    def __init__(self, d, n, h, look_ahead_mask = False, **kwargs):
        """
        a multi head self attention layer
            d    (int)----> number of features in the word embedding
            n    (int)----> number of words in the sentence (maximum length)
            h    (int)----> number of heads
            look_ahead_mask (bool)----> if True the model will not be able to see the future words
            used in the decoder part of the transformer model
        for more information about the parameters check the paper "Attention is all you need"
        """
        self.d   = d
        self.n   = n
        self.h   = h
        self.k   = self.d / self.h
        if look_ahead_mask:
            self.look_ahead_mask = (tf.linalg.band_part(tf.ones((self.n, self.n)), -1, 0) - 1) * 30e30
        else:
            self.look_ahead_mask = tf.zeros((self.n, self.n))
        super(Multi_Head_Attention, self).__init__(**kwargs)
    def build(self, input_shape):
        # intialize weights randomly using uniform distribution
        # Q, K, V, W are the weights of the model
        # all of the weights are expanded to enable batch maltiplication with the input matrix
        U_q = tf.random.uniform((self.h, self.d, self.d), minval = 0, maxval = 1)
        U_q = tf.expand_dims(U_q, 0)
        U_k = tf.random.uniform((self.h, self.d, self.d), minval = 0, maxval = 1)
        U_k = tf.expand_dims(U_k, 0)
        U_v = tf.random.uniform((self.h, self.d, self.d), minval = 0, maxval = 1)
        U_v = tf.expand_dims(U_v, 0)
        W   = tf.random.uniform((self.d, self.h * self.d), minval = 0, maxval = 1)
        W   = tf.expand_dims(W, 0)
        self.U_q = self.add_weight(shape = U_q.shape,\
                                initializer = tf.keras.initializers.Constant(U_q),\
                                trainable = True,\
                                dtype = tf.float32)
        self.U_k = self.add_weight(shape = U_k.shape,\
                                initializer = tf.keras.initializers.Constant(U_k),\
                                trainable = True,\
                                dtype = tf.float32)
        self.U_v = self.add_weight(shape = U_v.shape,\
                                initializer = tf.keras.initializers.Constant(U_v),\
                                trainable = True,\
                                dtype = tf.float32)
        self.W = self.add_weight(shape = W.shape,\
                                initializer = tf.keras.initializers.Constant(W),\
                                trainable = True,\
                                dtype = tf.float32)
    def call(self, x, encoder_inputs = False):
        """
        this function return multi head self attention tensor given the parameters and input matrix X
        inputs :
            X   (Tensor)-------------> input matrix of size [d,n]
            endocer_inputs (bool)----> if True the input is copied from the encoder and the decoder
        output :
            mhsa (Tensor)------------> matrix of size [d,n]
        """ 
        if encoder_inputs:
            # encoder output
            x1 = x[0]
            # decoder input
            x2 = x[1]
        else:
            x1 = x
            x2 = x
        batch_size =  tf.shape(x1)[0]
        # calculate the mask to prevent the paddings to slow down the model
        # get boolean vector to indicate position of padding  
        padding_mask = tf.math.not_equal(x1, 0)
        padding_mask = tf.reduce_all(padding_mask, axis = 1)
        # cast the vector from boolean to float32 then rshape it as column vector
        padding_mask = tf.cast(padding_mask, tf.float32)
        padding_mask = tf.reshape(padding_mask, (batch_size, 1, -1))
        # repeat that column n times 
        padding_mask = tf.concat([padding_mask]*self.n, axis = 1)
        padding_mask = (padding_mask - 1) * 30e30
        # add the look-ahead-mask
        mask = self.look_ahead_mask + padding_mask
        mask = tf.expand_dims(mask, 1)
        # caculate the multi-head attention
        x1   = tf.expand_dims(x1, 1)
        x2   = tf.expand_dims(x2, 1)
        Q    = tf.matmul(self.U_q, x2)  
        K    = tf.matmul(self.U_k, x1)
        V    = tf.matmul(self.U_v, x1)
        A    = tf.nn.softmax((tf.matmul(Q, K, transpose_a = True)  /tf.math.sqrt(self.k)) + mask ,\
                            axis = -1)
        Y    = V @ A
        mhsa = self.W @ tf.reshape(Y, [batch_size, self.h* self.d, self.n])
        return mhsa


import os
import pickle
import numpy as np
from nltk import everygrams
class TextMatcher:
    """
    TextMatcher is a module designed to address a common problem in Arabic NLP:
    Suppose you’re building a model to handle customer orders and process transactions automatically.
    You've implemented a Named Entity Recognition (NER) model to extract device names in Arabic,
    such as (غسالة, تلفاز, شاشة, مروحة).
    After identifying these devices, you need to match them with entries in your database to check stock availability.
    However, issues arise when there's a spelling variation.
    For instance, a customer might type "غساله," but in your database, the device is listed as "غسالة." 
    A standard search would fail to match these two, even though they represent the same item.
    This issue is known as orthographic or spelling variation and is a challenge for information retrieval in Arabic.
    Many approaches exist to address this, and in the ArabicTagger package, I approached the problem with a learnable weighted similarity.
    This method reduces the number of parameters and shortens training time while improving matching accuracy across spelling variations.
    for more information about the math behind the model
    see readme file in the github repository
    """
    def __init__(self, transformation_fun):
        """
        inputs:
        transformation_fun (fun)----> a function that is defined in [0,inf)
        this function is used to transform the position of the character in the word
        """
        self.transformation_fun = transformation_fun
        current_dir = os.path.dirname(__file__)
        directory   = os.path.join(current_dir, 'dict', "alphabets_indx.pkl")
        with open(directory,"rb")as file:
            self.alphabets_indx = pickle.load(file)
    def cosin_similiraty(self, x, y):
        # private function to calculate the cosine similarity between two vectors
        denominator = np.linalg.norm(x, axis = 0) * np.linalg.norm(y, axis = 0)
        denominator = denominator + 0.00001
        numerator   = np.sum(x * y, axis = 0) 
        return numerator/denominator
    def vectorize(self, words):
        # private function to vectorize the words
        # intialize a matrix as list of lists each list is a vector for a word
        # vector lenght is m = 1,260 
        matrix = []
        for word in words:
            vect = np.repeat(0.0001, len(self.alphabets_indx))
            # get every 1-gram and 2-gram of the word
            word_ng = everygrams(word, 1, 2)
            j = 0
            # the position of a single character is it's absolute position in the word
            # the position of a 2-gram is the average of the two characters positions
            for i,c in enumerate(word_ng):
                position = self.alphabets_indx.get(''.join(c))
                if (i+1)%2 == 0:
                    i = ((i-j) + (i-j+1))/2
                else:
                    i  = i - j
                    j += 1
                if position:
                    vect[position] += 1
                    vect[position] += self.transformation_fun(i)
            matrix.append(vect)
        X = np.array(matrix)
        return X
    
    def load_model(self, path = None):
        """
        load a pre-trained model or a model that was saved before
        input:
            path (str)-----> the path to the model must be the full path + name of file without any extension
            if the path is None a pre-trained model will be loaded 
        """
        if path:
            with open(str(path)+".pkl", 'rb') as file:
                self.W = pickle.load(file)
        else:
            current_dir = os.path.dirname(__file__)
            directory   = os.path.join(current_dir, 'models', "Matcher_weights1.pkl")
            with open(directory, 'rb') as file:
                container = pickle.load(file)
            self.W = container["W"]
            self.C = container["C"]
    def train(self):
        """
        the train method is used to train the model,must be called after fit method
        output:
            S_k (np.array)----> the similarity matrix for the positive class given the trained model
            S_j (np.array)----> the similarity matrix for the negative class given the trained model
            l_t (np.array)----> the labels for the positive class
            l_f (np.array)----> the labels for the negative class
        the outputs is used to calculate the optimal threshold C for the model and other metrics
        for more examples see kaggle notebook at github repository
        """
        w1_t, w1_f = self.train_data.query("l == True").w1, self.train_data.query("l == False").w1
        w2_t, w2_f = self.train_data.query("l == True").w2, self.train_data.query("l == False").w2
        l_t , l_f  = self.train_data.query("l == True").l, self.train_data.query("l == False").l
        I_k  = self.vectorize(w1_t)
        I_j  = self.vectorize(w1_f) 
        I_k_ = self.vectorize(w2_t)
        I_j_ = self.vectorize(w2_f)
        N_f  = 1 / (np.sum(self.cosin_similiraty(I_k , I_k_)) - \
                  np.sum(self.cosin_similiraty(I_j , I_j_)))
        W = N_f * (self.cosin_similiraty(I_k , I_k_) - self.cosin_similiraty(I_j , I_j_))
        W = W.reshape(-1,1)
        self.W = W
        S_k = (I_k * I_k_) @ (W * W)
        S_j = (I_j * I_j_) @ (W * W)
        return S_k,S_j,l_t,l_f
    
    def test(self):
        """
        the test method is used to test the model after training,
        must be called after train method
        output:
            S_k (np.array)----> the similarity matrix for the positive class given the trained model
            S_j (np.array)----> the similarity matrix for the negative class given the trained model
            l_t (np.array)----> the labels for the positive class
            l_f (np.array)----> the labels for the negative class
        the outputs is used to evaluate the model using the optimal threshold C on some metrics
        for more examples see kaggle notebook at github repository
        """
        w1_t, w1_f = self.test_data.query("l == True").w1, self.test_data.query("l == False").w1
        w2_t, w2_f = self.test_data.query("l == True").w2, self.test_data.query("l == False").w2
        l_t , l_f  = self.test_data.query("l == True").l, self.test_data.query("l == False").l
        I_k  = self.vectorize(w1_t)
        I_j  = self.vectorize(w1_f) 
        I_k_ = self.vectorize(w2_t)
        I_j_ = self.vectorize(w2_f)
        S_k  = (I_k * I_k_) @ (self.W * self.W)
        S_j  = (I_j * I_j_) @ (self.W * self.W)
        return S_k,S_j,l_t,l_f
    def predict(self, w1, w2, C):
        """
        predict method is used to predict the labels of the words
        give it two words that you want to compare and the optimal threshold C
        if the model is trained well it should predict the correct label
        1 means they are similar 0 means they are not
        inputs:
            w1 (List[str])----> the first word
            w2 (List[str])----> the second word
            C  (float)--------> the optimal threshold
        output:
            y_hat (np.array)--> the predicted labels
        """
        I_1   = self.vectorize(w1)
        I_2   = self.vectorize(w2) 
        S     = (I_1 * I_2) @ (self.W * self.W)
        y_hat = (S >= C) * 1
        return y_hat
    def fit(self, data, s):
        """
        fit method is used to split the data into training and testing sets 
        to prepare the data for training and testing the model
        inputs:
            data (pd.DataFrame)----> the data to split
            s    (float)-----------> the fraction of the data to be used for training s must be in (0,1)
        """
        data  = data.sample(frac = 1)
        s     = int(round(s * len(data)))
        self.train_data = data.iloc[:s,:]
        self.test_data  = data.iloc[s:,:]


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read(9591)


setup(
    name="ArabicTagger",  
    version="0.1.1b3",
    packages=find_packages(),
    install_requires=["tensorflow==2.10.0",\
                      "keras==2.10.0",\
                      "nltk",\
                       "bpemb", "numpy"],
    include_package_data=True,
    author="Eslam Tarek Farouk",
    author_email='cds.eslamtarek96337@alexu.edu.eg',
    description="A CRF layer and BI-LSTM+CRF model implemented in Keras",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EslamTarekFarouk/ArabicTagger",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
